/** @file fifo_model.cpp
 * This file contains the reference model class implementation.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <fifo_model.h>

namespace cpptesk {
namespace examples {
namespace fifo {

FIFO::FIFO() : coverage(*this)
{
    // Registration of interfaces
    ADD_INPUT(iface1);
    ADD_INPUT(iface2);
    ADD_OUTPUT(iface3);
}

FIFO::~FIFO() {}

bool FIFO::isIface1Ready() const
{
    // Return true if iface1 is free
    return iface1.isReady();
}

bool FIFO::isIface2Ready() const
{
    // Return true if iface2 is free
    return iface2.isReady();
}

DEFINE_STIMULUS(FIFO::push_msg)
{
    // Copy message to the local variable
    InputData data = CAST_MESSAGE(InputData);

    // Refresh FIFO model
    push_item(data.get_data());

    // Send message to the selected while calling FIFO::push_msg input interface in parallel mode
    START_STIMULUS(PARALLEL);

    // Wait for the next cycle of simulation
    CYCLE();

    // Stop processing this stimulus
    STOP_STIMULUS();
}

DEFINE_STIMULUS(FIFO::pop_msg)
{
    // Copy message to the local variable
    InputData data = CAST_MESSAGE(InputData);

    // Start processing this stimulus
    START_PROCESS();

    // Send message "data" to the standard interface ("GET_IFACE()" corresponds to the selected while calling FIFO::pop_msg interface) in parallel mode
    RECV_STIMULUS(PARALLEL, GET_IFACE(), data);

    // Preparation of the message to the output interface
    OutputData outdata = OutputData(pop_item());

    // Start reaction "receive data sent as the result of this stimulus"
    get_pop_msg(process, iface3, outdata);

    // Stop processing this stimulus
    STOP_PROCESS();
}

DEFINE_REACTION(FIFO::get_pop_msg)
{
    START_PROCESS();
    // Send reaction to the adapter (FET_IFACE() and GET_MESSAGE() mean that interface and message have been selected while calling FIFO::get_pop_msg)
    SEND_REACTION(SEQUENTIAL, GET_IFACE(), GET_MESSAGE());

    STOP_PROCESS();
}

void FIFO::push_item(uint8_t data)
{
    // Starting writting is prohibited if FIFO is full
    assert(fifo.size() < 16);
    // Push data into FIFO reference model
    fifo.push_back(data);
}

uint8_t FIFO::pop_item(void)
{
    // Starting reading is prohibited if FIFO is empty
    assert(fifo.size() > 0);
    // Data on the FIFO top
    uint8_t data = fifo[0];
    // Delete this read data
    fifo.erase(fifo.begin());
    // Return the data
    return data;
}

bool FIFO::is_empty(void)
{
    return fifo.size() == 0;
}

bool FIFO::is_full(void)
{
    return fifo.size() == 16;
}

int FIFO::fifo_fullness(void)
{
    return fifo.size();
}

void FIFO::refresh_coverage(void)
{
    coverage.out_cov_FF();
}

}}}

