/** @file fifo_media.cpp
 * Mediator for united FIFO model.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <multififo_media.h>

#include <sync.h>

#include <interface.h>
#include <vpi_media.h>

#include <iostream>

namespace cpptesk {
namespace examples {
namespace fifo {

using namespace cpptesk::hw;

MultififoMediator::MultififoMediator()
{
    // Set level of debug messages
    SET_DEBUG_LEVEL(DEBUG_INFO, true);
    SET_DEBUG_STYLE(DEBUG_USER, BLACK, CYAN, CYAN, BLACK);

    // Registration of correspondence between serializers, deserializers and ifaces.
    SET_INPUT_ADAPTER(iface1a, MultififoMediator::serialize_iface1a);
    SET_INPUT_ADAPTER(iface1b, MultififoMediator::serialize_iface1b);
    SET_INPUT_ADAPTER(iface2, MultififoMediator::serialize_iface2);
    SET_OUTPUT_ADAPTER(iface3, MultififoMediator::deserialize_iface3);

    CPPTESK_SET_OUTPUT_ADAPTER_OF(FIFOMediator, fifo0, fifo0.iface3, MultififoMediator::deserialize_inner_iface0);
    CPPTESK_SET_OUTPUT_ADAPTER_OF(FIFOMediator, fifo1, fifo1.iface3, MultififoMediator::deserialize_inner_iface1);

    // Set maximum timeout.
    setReactionTimeout(MultififoMediator::REACTION_TIMEOUT);

    // Registration of arbiter of output iface3.
    SET_ARBITER(iface3, matcher_iface3);

    // Detectors of extra reactions should be started on the first cycle.
    CALL_OUTPUT_LISTENER(MultififoMediator::listen_iface3, iface3);
}

MultififoMediator::~MultififoMediator() {}

bool MultififoMediator::isIface1aReady() const
{
    return Multififo::isIface1aReady();
}
bool MultififoMediator::isIface1bReady() const
{
    return Multififo::isIface1bReady();
}
bool MultififoMediator::isIface2Ready() const
{
    return Multififo::isIface2Ready();
}

CPPTESK_DEFINE_STIMULUS(MultififoMediator::serialize_iface1a)
{
    // Copy data to the local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start of serialization processing.
    START_PROCESS();

    DEBUG_PRINT(DEBUG_USER, "MultififoMediator::serialize_iface1a");

    // Capture iface1a.
    CAPTURE_IFACE();

    // Set data to the wires of DUT.
    inputs.CE_WR0 = 1;
    inputs.DI0 = msg.get_data();

    // Release of iface1a.
    RELEASE_IFACE();

    // Wait one cycle of simulation.
    CYCLE();

    // Finish of serialization processing.
    STOP_PROCESS();
}

CPPTESK_DEFINE_STIMULUS(MultififoMediator::serialize_iface1b)
{
    // Copy data to the local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start of serialization processing.
    START_PROCESS();

    DEBUG_PRINT(DEBUG_USER, "MultififoMediator::serialize_iface1b");

    // Capture iface1b.
    CAPTURE_IFACE();

    // Set data to the wires of DUT.
    inputs.CE_WR1 = 1;
    inputs.DI1 = msg.get_data();

    // Release of iface1b.
    RELEASE_IFACE();

    // Wait one cycle of simulation.
    CYCLE();

    // Finish of serialization processing.
    STOP_PROCESS();
}

CPPTESK_DEFINE_STIMULUS(MultififoMediator::serialize_iface2)
{
    // Copy data to the local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start of serialization processing.
    START_PROCESS();

    DEBUG_PRINT(DEBUG_USER, "MultififoMediator::serialize_iface2");

    // Capture iface2.
    CAPTURE_IFACE();

    // Set data to the wires of DUT.
    inputs.CE_RD = 1;

    // Release of iface2.
    RELEASE_IFACE();

    // Wait one cycle of simulation.
    CYCLE();

    // Finish of serialization processing.
    STOP_PROCESS();
}

CPPTESK_DEFINE_REACTION(MultififoMediator::deserialize_iface3)
{
    // Copy data to the local variable.
    OutputData &msg = CPPTESK_CAST_MESSAGE(OutputData);

    // Start of deserialization processing.
    START_PROCESS();

    // Wainting for reaction by means of DUT's DO_VAL wire.
    WAIT_REACTION(outputs.DO_VAL);

    // Output data are saved to the following check.
    msg.set_data(outputs.DO);

    // Jump to the next registered reaction (if it exists).
    NEXT_REACTION();

    // Finish of deserialization processing.
    STOP_PROCESS();
}

CPPTESK_DEFINE_PROCESS(MultififoMediator::deserialize_inner_iface0)
{
    OutputData msg = CPPTESK_CAST_MESSAGE(OutputData);

    START_PROCESS();

    DEBUG_PRINT(DEBUG_USER, "MultififoMediator::deserialize_inner_iface0");
    DEBUG_PRINTF(DEBUG_USER, "iface0: %llx\n", msg.get_data());

    CAPTURE_IFACE();

    if(!fifo2.is_full()) {
        InputData data = InputData();
        data.set_data(msg.get_data());
        fifo2.start(&FIFO::push_msg, fifo2.iface1, data, FIFO::START_REQUEST_IFACE);
    } else { DEBUG_PRINT(DEBUG_USER, "but the fifo2 is full"); }

    RELEASE_IFACE();

    CYCLE_OF(fifo0);

    STOP_PROCESS();
}

CPPTESK_DEFINE_PROCESS(MultififoMediator::deserialize_inner_iface1)
{
    OutputData msg = CPPTESK_CAST_MESSAGE(OutputData);

    START_PROCESS();

    DEBUG_PRINT(DEBUG_USER, "MultififoMediator::deserialize_inner_iface1");
    DEBUG_PRINTF(DEBUG_USER, "iface1: %llx\n", msg.get_data());

    CAPTURE_IFACE();

    if(!fifo2.is_full()) {
        while(!fifo2.isIface1Ready()) CYCLE();
        InputData data = InputData();
        data.set_data(msg.get_data());
        fifo2.start(&FIFO::push_msg, fifo2.iface1, data, FIFO::START_REQUEST_IFACE);
    } else { DEBUG_PRINT(DEBUG_USER, "but the fifo2 is full"); }

    RELEASE_IFACE();

    CYCLE_OF(fifo1);

    STOP();
}

void MultififoMediator::cycle()
{
    fifo0.cycle();
    fifo1.cycle();
    fifo2.cycle();
    Multififo::cycle();
}

}}}
