/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Point event is a triple <message, interface, time>.
//           Prolonged event has an additional time dimension (duration).
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/errors.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>
#include <hw/time.hpp>
#include <hw/utils/shared_ptr.hpp>

#include <string>
#include <sstream>

namespace cpptesk {
namespace hw {

class PointEvent 
{
public:
    PointEvent(const Interface &iface, const Message &message, cycle_t start_time):
        _iface(&iface), _message(message.clone()), _start_time(start_time) {}

    PointEvent(const Interface &iface, cycle_t start_time):
        _iface(&iface), _message(0), _start_time(start_time) {}

    virtual ~PointEvent() {}

    inline const Interface *where() const { return _iface; }
    inline const Message *what() const { return _message.get(); }
    inline cycle_t when() const { return _start_time; }

protected:
    const Interface *_iface;
    SharedPtr<Message> _message;
    cycle_t _start_time;
};

class ProlongedEvent: public PointEvent
{
public:
    ProlongedEvent(const Interface &iface, const Message &message, cycle_t start_time, cycle_t stop_time):
        PointEvent(iface, message, start_time), _stop_time(stop_time) {}

    ProlongedEvent(const Interface &iface, cycle_t start_time, cycle_t stop_time):
        PointEvent(iface, start_time), _stop_time(stop_time) {}

    virtual ~ProlongedEvent() {}

    inline cycle_t whenStarted() const { return _start_time; }
    inline cycle_t whenStopped() const { return _stop_time; }

protected:
    cycle_t _stop_time;
};

class Stimulus: public ProlongedEvent 
{
public:
    Stimulus(const Interface &iface, const Message &message, cycle_t start_time, cycle_t stop_time):
        ProlongedEvent(iface, message, start_time, stop_time) {}

    Stimulus(const Interface &iface, cycle_t start_time, cycle_t stop_time):
        ProlongedEvent(iface, start_time, stop_time) {}

    virtual ~Stimulus() {}
};

class Reaction: public ProlongedEvent 
{
public:
    Reaction(error_type_t error, const Interface &iface, const Message &spec_msg, const Message &impl_msg,
        cycle_t start_time, cycle_t stop_time, const std::string &info, unsigned type = -1u):
        ProlongedEvent(iface, spec_msg, start_time, stop_time), _impl_msg(impl_msg.clone()), _error(error), _info(info), _type(type) {}

    Reaction(error_type_t error, const Interface &iface, const Message &spec_msg,
        cycle_t start_time, cycle_t stop_time, const std::string &info, unsigned type = -1u):
	ProlongedEvent(iface, spec_msg, start_time, stop_time), _impl_msg(0), _error(error), _info(info), _type(type) {}

    Reaction(error_type_t error, const Interface &iface,
        cycle_t start_time, cycle_t stop_time, const std::string &info, unsigned type = -1u):
        ProlongedEvent(iface, start_time, stop_time), _impl_msg(0), _error(error), _info(info), _type(type) {}

    virtual ~Reaction() {}

    const Message *getSpecMessage() const { return _message.get();  }
    const Message *getImplMessage() const { return _impl_msg.get(); }

    /// Returns true iff the reaction is erroneous.
    bool isError() const { return _error != NORMAL_REACTION; }

    /// Returns the error type.
    error_type_t getError() const { return _error; }

    /// Returns the reaction's type (@depricated).
    uint32_t getType() const { return _type; }

    /// Returns the reaction's info.
    const std::string& getInfo() const { return _info; }

    /// Returns short description of the reaction.
    std::string getShortDescription() const
    {
        std::stringstream buffer;
        buffer << getErrorString(_error) << "(iface=" << *_iface << " cycle=" << std::dec << _start_time << ")";
        return buffer.str();
    }

    /// Returns full description of the reaction.
    std::string getFullDescription() const
    {
        std::stringstream buffer;

        if(_info.empty())
	    { return getShortDescription(); }

        buffer << getShortDescription() << ": " << _info;
        return buffer.str();
    }

protected:
    SharedPtr<Message> _impl_msg;
    error_type_t _error;
    std::string _info;
    unsigned _type;
};

inline std::ostream& operator << (std::ostream& out, const Reaction &reaction)
{
    return out << reaction.getFullDescription();
}

}} // namespace cpptesk::hw

