/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: History of events (stimuli & reactions) 
///////////////////////////////////////////////////////////////////////////////

#include <hw/history.hpp>
#include <hw/unused.hpp>

#include <iomanip>
#include <sstream>

#include <string.h>

#ifdef CPPTESK_BUILD_WITH_OLD_ASPECTRACE_PATH
extern "C" {
#include <aspectrace.h>
#include <aspect_coverage.h>
#include <aspect_error.h>
}
#else
extern "C" {
#include <aspectrace/aspectrace.h>
#include <aspectrace/aspect_coverage.h>
#include <aspectrace/aspect_error.h>
}
#endif // CPPTESK_BUILD_WITH_OLD_ASPECTRACE_PATH

namespace cpptesk {
namespace hw {

EventHistory::EventHistory()
{
    reset();
}

EventHistory::~EventHistory()
{
    reset();
}

void EventHistory::registerStimulus(const Stimulus &stimulus)
{
    CPPTESK_UNUSED_ARG(stimulus);

    _stimulus_count++;
}

void EventHistory::registerReaction(const Reaction &reaction)
{
    _reactions.push_back(reaction);
    _reaction_count[reaction.getError()]++;
}

void EventHistory::clearReactions()
{
    // Append the current set to the history
    for(reaction_list_t::iterator i = _reactions.begin(); i != _reactions.end(); i++)
        { _history.push_back(*i); }

    // Reduce the history
    if(_history.size() > HISTORY_SIZE)
        { _history.erase(_history.begin(), _history.begin() + (_history.size() - HISTORY_SIZE)); }

    // Clear the current set
    _reactions.clear();
}

const std::vector<Reaction>& EventHistory::getReactions() const
{
    return _reactions;
}

bool EventHistory::verdict()
{
    bool result = true;

    // If all reactions are normal, then the verdict is true
    for(reaction_list_t::iterator i = _reactions.begin(); i != _reactions.end(); i++)
    {
        // Otherwise, the verdict is false
        if(i->getError() != NORMAL_REACTION)
        { 
            result = false; 
            aspect_error_traceFailure("error", i->getShortDescription().c_str(), i->getFullDescription().c_str()); 
        }
    }

    // The reactions have been analyzed and there is no reason to analyze them once again
    clearReactions();

    return result;
}

std::string EventHistory::error(error_mode_t mode) const
{
    bool written  = false;
    bool finished = false;
    std::stringstream buffer;

    // Analyze the current set of reactions
    for(reaction_list_t::const_reverse_iterator i = _reactions.rbegin(); i != _reactions.rend(); i++)
    {
        if(i->getError() != NORMAL_REACTION)
        {
            if(written) { buffer << std::endl; }
            buffer << i->getFullDescription();
            if(mode == LAST_ERROR) { finished = true; break; }
            written = true;
        }

        if(mode == LAST_REACTION) { finished = true; break; }
    }

    if(finished)
        { return buffer.str(); }

    // Analyze the history of reactions
    for(reaction_list_t::const_reverse_iterator i = _history.rbegin(); i != _history.rend(); i++)
    {
        if(i->getError() != NORMAL_REACTION)
        {
            if(written) { buffer << std::endl; }
            buffer << i->getFullDescription();
            if(mode == LAST_ERROR) { break; }
            written = true;
        }

        if(mode == LAST_REACTION) { break; }
    }

    return buffer.str();
}

std::string EventHistory::diagnosis() const
{
    // TODO: Analyze the history only
    return "";
}

unsigned EventHistory::count(error_type_t error) const
{
    return _reaction_count[error];
}

void EventHistory::reset()
{
    // Reset the current set of reactions
    _reactions.clear();
    // Reset the history of reactions
    _history.clear();
    // Reset statistics on stimuli
    _stimulus_count = 0;
    // Reset statistics on reactions
    ::memset(_reaction_count, 0, sizeof(_reaction_count));
}

void EventHistory::traceStatistics()
{
    unsigned max_width = 0;
    unsigned all_count = 0;

    std::stringstream buffer;

    for(unsigned error = 0; error < __ERROR_TYPE_NUMBER; error++)
    {
       const std::string &error_string = getErrorString((error_type_t)error);
       if(error_string.length() > max_width) { max_width = error_string.length(); }
       all_count += count((error_type_t)error);
    }

    buffer << "Statistics on stimuli and reactions:" << std::endl;

    buffer.width(max_width); buffer.fill(' ');
    buffer << std::left << "[STIMULUS]" << ": ";
    buffer << std::dec << _stimulus_count << " (" << std::fixed << std::setprecision(2)
        << (all_count > 0 ? (float)_stimulus_count/(float)all_count : 0.0) << " S/R)" << std::endl;

    buffer.width(max_width);
    buffer << std::left << "[REACTION]" << ": ";
    buffer << std::dec << all_count << " (" << std::fixed << std::setprecision(2)
        << (_stimulus_count > 0 ? (float)all_count/(float)_stimulus_count : 0.0) << " R/S)";

    for(unsigned error = 0; error < __ERROR_TYPE_NUMBER; error++)
    {
        buffer.width(max_width);
	buffer << std::endl;
        buffer << std::left << getErrorString((error_type_t)error) << ": ";
        buffer << std::dec << count((error_type_t)error);
    }

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, buffer.str());
}

}} // namespace cpptesk::hw
