/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Basic model of a synchronous hardware design
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/component.hpp>
#include <hw/debug.hpp>
#include <hw/history.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>
#include <hw/process.hpp>
#include <hw/random.hpp>
#include <hw/signals.hpp>
#include <hw/stack.hpp>
#include <hw/synchronizer.hpp>
#include <hw/time.hpp>
#include <hw/unused.hpp>
#include <hw/arbiter/arbiter.hpp>
#include <hw/sequence/manager.hpp>

#include <algorithm>
#include <map>
#include <iostream>
#include <sstream>
#include <list>
#include <vector>

#include <assert.h>
#include <stdint.h>

#if defined(WIN32)
    #include <hw/utils/win32/ucontext.h>
#elif defined(__GNUC__)
    #include <ucontext.h>
#else
    #error "unknown platform"
#endif

////////////////////////////////////////////////////////////////////////////////
// Return Types of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_PROCESS
 * Keywork marking a model process.
 */
#define CPPTESK_PROCESS void

/** @def CPPTESK_STIMULUS
 * Keywork marking a stimulus.
 */
#define CPPTESK_STIMULUS CPPTESK_PROCESS

/** @def CPPTESK_REACTION
 * Keywork marking a reaction.
 */
#define CPPTESK_REACTION CPPTESK_PROCESS

#ifdef HW_SHORT_NAMES
    #define PROCESS  CPPTESK_PROCESS
    #define STIMULUS CPPTESK_STIMULUS
    #define REACTION CPPTESK_REACTION
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Model Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MODEL
 * Declares a model of a hardware component.
 */
#define CPPTESK_MODEL(UserModel) \
    struct UserModel: public ::cpptesk::hw::Model<UserModel>

#define CPPTESK_SIGNAL_MODEL(UserModel, Inputs, Outputs) \
    struct UserModel: public ::cpptesk::hw::Model<UserModel>, public ::cpptesk::hw::IOSignals<Inputs, Outputs>

#ifdef HW_SHORT_NAMES
    #define MODEL(UserModel)                         CPPTESK_MODEL(UserModel)
    #define SIGNAL_MODEL(UserModel, Inputs, Outputs) CPPTESK_SIGNAL_MODEL(UserModel, Inputs, Outputs)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Parameters of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GET_PROCESS
 * Returns a reference to a process parameter.
 */
#define CPPTESK_GET_PROCESS() process

/** @def CPPTESK_GET_IFACE
 * Returns a reference to an interface parameter.
 */
#define CPPTESK_GET_IFACE() iface

/** @def CPPTESK_GET_MESSAGE
 * Returns a reference to a message parameter.
 */
#define CPPTESK_GET_MESSAGE() message

/** @def CPPTESK_CAST_MESSAGE
 * Casts a message parameter to a given type and returns a reference.
 */
#define CPPTESK_CAST_MESSAGE(MessageType) \
    ::cpptesk::hw::_cpptesk_cast_message<MessageType>(CPPTESK_GET_MESSAGE(), __FILE__, __LINE__)

#define CPPTESK_CONST_CAST_MESSAGE(MessageType) \
    ::cpptesk::hw::_cpptesk_const_cast_message<MessageType>(CPPTESK_GET_MESSAGE(), __FILE__, __LINE__)

/** @def CPPTESK_GET_PARAMETERS
 * Returns a comma-separated list of parameters.
 */
#define CPPTESK_GET_PARAMETERS() \
    CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE()

#ifdef HW_SHORT_NAMES
    #define GET_PROCESS()                   CPPTESK_GET_PROCESS()
    #define GET_IFACE()                     CPPTESK_GET_IFACE()
    #define GET_MESSAGE()                   CPPTESK_GET_MESSAGE()
    #define CAST_MESSAGE(MessageType)       CPPTESK_CAST_MESSAGE(MessageType)
    #define CONST_CAST_MESSAGE(MessageType) CPPTESK_CONST_CAST_MESSAGE(MessageType)
    #define GET_PARAMETERS()                CPPTESK_GET_PARAMETERS()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DECLARE_PROCESS
 * Declares a process within a model class.
 */
#define CPPTESK_DECLARE_PROCESS(process_name) \
    virtual CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_NON_VIRTUAL_PROCESS
 * Declares a non-virtal process within a model class.
 */
#define CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(process_name) \
    CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_STIMULUS
 * Declares a stimulus within a model class.
 */
#define CPPTESK_DECLARE_STIMULUS(stimulus_name) \
    virtual CPPTESK_STIMULUS stimulus_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_PRECONDITION
 * Declares a stimulus precondition within a model class.
 */
#define CPPTESK_DECLARE_PRECONDITION(stimulus_name) \
    virtual bool pre_##stimulus_name(::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), \
        ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_STIMULUS_WITH_PRE \
 * Declares a stimulus and its precondition.
 */
#define CPPTESK_DECLARE_STIMULUS_WITH_PRE(stimulus_name) \
    CPPTESK_DECLARE_PRECONDITION(stimulus_name); \
    CPPTESK_DECLARE_STIMULUS(stimulus_name)

/** @def CPPTESK_DECLARE_REACTION
 * Declares a reaction within a model class.
 */
#define CPPTESK_DECLARE_REACTION(reaction_name) \
    virtual CPPTESK_REACTION reaction_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

#ifdef HW_SHORT_NAMES
    #define DECLARE_PROCESS(process_name)             CPPTESK_DECLARE_PROCESS(process_name)
    #define DECLARE_NON_VIRTUAL_PROCESS(process_name) CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(process_name)
    #define DECLARE_STIMULUS(stimulus_name)           CPPTESK_DECLARE_STIMULUS(stimulus_name)
    #define DECLARE_PRECONDITION(stimulus_name)       CPPTESK_DECLARE_PRECONDITION(stimulus_name)
    #define DECLARE_STIMULUS_WITH_PRE(stimulus_name)  CPPTESK_DECLARE_STIMULUS_WITH_PRE(stimulus_name)
    #define DECLARE_REACTION(reaction_name)           CPPTESK_DECLARE_REACTION(reaction_name)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Special-Purpose Fields
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DECLARE_IFACE
 * Declares a model interface.
 */
#define CPPTESK_DECLARE_IFACE(iface) \
    ::cpptesk::hw::Interface iface

/** @def CPPTESK_DECLARE_INPUT
 * Declares an input interface.
 */
#define CPPTESK_DECLARE_INPUT(iface) \
    ::cpptesk::hw::InputInterface iface

/** @def CPPTESK_DECLARE_OUTPUT
 * Declares an output interface.
 */
#define CPPTESK_DECLARE_OUTPUT(iface) \
    ::cpptesk::hw::OutputInterface iface

/** @def CPPTESK_DECLARE_INOUT
 * Declares an input/output interface.
 */
#define CPPTESK_DECLARE_INOUT(iface) \
    ::cpptesk::hw::InoutInterface iface

/** @def CPPTESK_DECLARE_ARBITER
 * Declares a primary reaction arbiter (matcher).
 */
#define CPPTESK_DECLARE_ARBITER(type, arbiter) \
    ::cpptesk::hw::type##ReactionArbiter arbiter

/** @def CPPTESK_DECLARE_TWO_LEVEL_ARBITER
 * Declares two-level reaction arbiter (matcher).
 */
#define CPPTESK_DECLARE_TWO_LEVEL_ARBITER(A1, A2) \
    ::cpptesk::hw::TwoLevelArbiter arbiter(A1, A2)

#ifdef HW_SHORT_NAMES
    #define DECLARE_IFACE(iface)           CPPTESK_DECLARE_IFACE(iface)
    #define DECLARE_INPUT(iface)           CPPTESK_DECLARE_INPUT(iface)
    #define DECLARE_OUTPUT(iface)          CPPTESK_DECLARE_OUTPUT(iface)
    #define DECLARE_INOUT(iface)           CPPTESK_DECLARE_INOUT(iface)
    #define DECLARE_ARBITER(type, arbiter) CPPTESK_DECLARE_ARBITER(type, arbiter)
    #define DECLARE_TWO_LEVEL_ARBITER(A1, A2) CPPTESK_DECLARE_TWO_LEVEL_ARBITER(A1, A2)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Definition of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DEFINE_PROCESS
 * Defines a process.
 */
#define CPPTESK_DEFINE_PROCESS(process_name) \
    CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DEFINE_STIMULUS
 * Defines a stimulus.
 */
#define CPPTESK_DEFINE_STIMULUS(stimulus_name) \
    CPPTESK_STIMULUS stimulus_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DEFINE_PRECONDITION
 * Defines a stimulus precondition.
 */
#define CPPTESK_DEFINE_PRECONDITION(pre_name) \
    bool pre_name(::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), \
        ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DEFINE_REACTION
 * Defines a reaction within a model class.
 */
#define CPPTESK_DEFINE_REACTION(reaction_name) \
    CPPTESK_REACTION reaction_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

#ifdef HW_SHORT_NAMES
    #define DEFINE_PROCESS(process_name)   CPPTESK_DEFINE_PROCESS(process_name)
    #define DEFINE_STIMULUS(stimulus_name) CPPTESK_DEFINE_STIMULUS(stimulus_name)
    #define DEFINE_PRECONDITION(pre_name)  CPPTESK_DEFINE_PRECONDITION(pre_name)
    #define DEFINE_REACTION(reaction_name) CPPTESK_DEFINE_REACTION(reaction_name)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Registration of Special-Purpose Components
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_ADD_COMPONENT
 * Registers a component in a model.
 */
#define CPPTESK_ADD_COMPONENT(component) \
    do { \
        CPPTESK_ASSERTION(!(component).getParent(), "Component " << #component << " has been already registered"); \
        (component).updateName(#component); \
        (component).setParent(this); \
        registerComponent(component); \
    } while(false)

/** @def CPPTESK_ADD_IFACE
 * Registers an interface in a model.
 */
#define CPPTESK_ADD_IFACE(iface) CPPTESK_ADD_COMPONENT(iface)

/** @def CPPTESK_ADD_INPUT
 * Registers an input interface in a model.
 */
#define CPPTESK_ADD_INPUT(iface) CPPTESK_ADD_COMPONENT(iface)

/** @def CPPTESK_ADD_OUTPUT
 * Registers an output interface in a model.
 */
#define CPPTESK_ADD_OUTPUT(iface) CPPTESK_ADD_COMPONENT(iface)

/** @def CPPTESK_SET_ARBITER
 * Registers a primary arbiter for a given output interface.
 */
#define CPPTESK_SET_ARBITER(iface, arbiter) \
    iface.registerReactionArbiter(arbiter)

#ifdef HW_SHORT_NAMES
    #define ADD_COMPONENT(component)    CPPTESK_ADD_COMPONENT(component)
    #define ADD_IFACE(iface)            CPPTESK_ADD_IFACE(iface)
    #define ADD_INPUT(iface)            CPPTESK_ADD_INPUT(iface)
    #define ADD_OUTPUT(iface)           CPPTESK_ADD_OUTPUT(iface)
    #define SET_ARBITER(iface, arbiter) CPPTESK_SET_ARBITER(iface, arbiter)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Parameters
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_SET_REACTION_TIMEOUT
 * Sets reaction timeout.
 */
#define CPPTESK_SET_REACTION_TIMEOUT(timeout) \
    setReactionTimeout(timeout)

#ifdef HW_SHORT_NAMES
    #define SET_REACTION_TIMEOUT(timeout) CPPTESK_SET_REACTION_TIMEOUT(timeout)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Control Statements
////////////////////////////////////////////////////////////////////////////////

/**
 * Process status is reassigned to EXECUTING to handle nested calls.
 */
#define __CPPTESK_SYNPOINT(type, name) \
    { \
        CPPTESK_GET_PROCESS().point = ::cpptesk::hw::Process::PointDescriptor(type, name, __FILE__, __LINE__); \
        /* IMPORTANT: Timeout flag is not reset by CYCLE() and DELAY() */ \
        if(type == ::cpptesk::hw::GUARD) { CPPTESK_GET_PROCESS().timeout = false; } \
        CPPTESK_GET_PROCESS().status = ::cpptesk::hw::EXECUTING; \
    }

#define __CPPTESK_RETURN(result) \
    do { \
        /* To check presense of CPPTESK_START_PROCESS() */ \
        __CPPTESK_CHECK_START(); \
        CPPTESK_ASSERTION(CPPTESK_GET_PROCESS().call_stack_size() > 0, "There is no an opening START for the current STOP"); \
        CPPTESK_GET_PROCESS().status = (result); \
        /* External parallel process jumps to the parent process */ \
        CPPTESK_GET_PROCESS().stop_call(); \
        if(CPPTESK_GET_PROCESS().call_stack_size() == 0) { \
            suspend(&CPPTESK_GET_PROCESS()); \
        } \
        return; \
    } while(false)

////////////////////////////////////////////////////////////////////////////////
// Precondition and Postcondition
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_CHECK_CONDITION
 * Returns false if a condition is not satistied. The macros is typically used
 * in precondition methods.
 */
#define CPPTESK_CHECK_CONDITION(cond) \
    if(!(cond)) \
    { \
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_INFO, "Condition " << #cond << " is violated"); \
        return false; \
    }

/** @def CPPTESK_PRE
 * Precondition assertion. The macros is typically used in stimulus methods.
 */
#define CPPTESK_PRE(cond) \
    if(!(cond)) \
    { \
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_INFO, "Precondition " << #cond << " is violated"); \
        CPPTESK_GET_PROCESS().point.file = __FILE__; \
        CPPTESK_GET_PROCESS().point.line = __LINE__; \
        __CPPTESK_RETURN(::cpptesk::hw::UNPREDICT); \
    }

/** @def CPPTESK_POST
 * Postcondition assertion. The macros is typically used in stimulus methods.
 */
#define CPPTESK_POST(cond) \
    if(!(cond)) \
    { \
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_INFO, "Postcondition " << #cond << " is violated"); \
        CPPTESK_GET_PROCESS().point.file = __FILE__; \
        CPPTESK_GET_PROCESS().point.line = __LINE__; \
        __CPPTESK_RETURN(::cpptesk::hw::VIOLATION); \
    }

#ifdef HW_SHORT_NAMES
    #define CHECK_CONDITION(cond) CPPTESK_CHECK_CONDITION(cond)
    #define PRE(cond)             CPPTESK_PRE(cond)
    #define POST(cond)            CPPTESK_POST(cond)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Call Process
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_PREPARE_CALL(method_name) \
    prepare_call(::cpptesk::hw::Process::CallDescriptor(#method_name, __FILE__, __LINE__))

#define CPPTESK_CALL_PARALLEL(method_name, iface, message) \
    ( \
        /* IMPORTANT: Do not check existance of START() */ \
        __CPPTESK_PREPARE_CALL(method_name), \
        start((model_t::process_t)&model_t::method_name, iface, message, START_IMMEDIATELY) \
    )

#define CPPTESK_CALL_SEQUENTIAL(method_name, iface, message) \
    ( \
        __CPPTESK_CHECK_START(), \
        __CPPTESK_PREPARE_CALL(method_name), \
        start((model_t::process_t)&model_t::method_name, CPPTESK_GET_PROCESS(), iface, message) \
    )

#define CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) \
    ( \
        /* IMPORTANT: Do not check existence of START() */ \
        __CPPTESK_PREPARE_CALL(full_method_name), \
        start((model_t::process_t)&full_method_name, iface, message, mode) \
    )

#define CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message) \
    ( \
        /* IMPORANT: Do not check existence of START() */ \
        __CPPTESK_PREPARE_CALL(full_method_name), \
        start((model_t::process_t)&full_method_name, CPPTESK_GET_PROCESS(), iface, message) \
    )

#define CPPTESK_CALL_PROCESS(call_mode, method_name, iface, message) \
    ((call_mode) == PARALLEL ? \
        CPPTESK_CALL_PARALLEL(method_name, iface, message) : \
        CPPTESK_CALL_SEQUENTIAL(method_name, iface, message))

#define CPPTESK_CALL_PROCESS_EX(call_mode, full_method_name, iface, message, mode) \
    ((call_mode) == PARALLEL ? \
        /* IMPORTANT: Parallel call does not use the process parameter */ \
        CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) : \
        /* IMPORTANT: Sequential call does not use the mode parameter */ \
        CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message, mode))

#define CPPTESK_RECV_STIMULUS(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onStimulus, iface, message)

#define CPPTESK_SEND_REACTION(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
   _cpptesk_set_explanation(message, time(), time() + getReactionTimeout(), __FILE__, __LINE__))

#define CPPTESK_SEND_REACTION_WITH_COMMENT(call_mode, iface, message, comment) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
    _cpptesk_set_explanation(message, time(), time() + getReactionTimeout(), __FILE__, __LINE__, comment))

#define CPPTESK_CALL_PROCESS_OF(model, full_method_name, iface, message) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        (model).start(&full_method_name, iface, message) \
    )

#define CPPTESK_CALL_STIMULUS_OF(model, full_method_name, iface, message) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        (model).start(&full_method_name, iface, message, ::cpptesk::hw::Model<cpptesk::hw::EMPTY_MODEL>::START_REQUEST_IFACE) \
    )

/** @def CPPTESK_CALL_LOOP
 * Launches an infinite loop.
 */
#define CPPTESK_CALL_LOOP(method_name) \
    CPPTESK_CALL_PARALLEL(method_name, CPPTESK_DEFAULT_IFACE, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE)

/** @def CPPTESK_CALL_LOOP_EX
 * Launches an infinite loop (usually, a listener of an output interface).
 */
#define CPPTESK_CALL_LOOP_EX(full_method_name, iface, mode) \
    CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE, mode)

#ifdef HW_SHORT_NAMES
    #define CALL_PARALLEL(method_name, iface, message)                         CPPTESK_CALL_PARALLEL(method_name, iface, message)
    #define CALL_SEQUENTIAL(method_name, iface, message)                       CPPTESK_CALL_SEQUENTIAL(method_name, iface, message)
    #define CALL_PROCESS(call_mode, method_name, iface, message)               CPPTESK_CALL_PROCESS(call_mode, method_name, iface, message)
    #define CALL_PARALLEL_EX(full_method_name, iface, message, mode)           CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode)
    #define CALL_SEQUENTIAL_EX(full_method_name, iface, message)               CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message)
    #define CALL_PROCESS_EX(call_mode, full_method_name, iface, message, mode) CPPTESK_CALL_PROCESS_EX(call_mode, full_method_name, iface, message, mode)
    #define RECV_STIMULUS(call_mode, iface, message)                           CPPTESK_RECV_STIMULUS(call_mode, iface, message)
    #define SEND_REACTION(call_mode, iface, message)                           CPPTESK_SEND_REACTION(call_mode, iface, message)
    #define SEND_REACTION_WITH_COMMENT(call_mode, iface, message, comment)     CPPTESK_SEND_REACTION_WITH_COMMENT(call_mode, iface, message, comment)
    #define CALL_PROCESS_OF(model, full_method_name, iface, message)           CPPTESK_CALL_PROCESS_OF(model, full_method_name, iface, message)
    #define CALL_STIMULUS_OF(model, full_method_name, iface, message)          CPPTESK_CALL_STIMULUS_OF(model, full_method_name, iface, message)
    #define CALL_LOOP(method_name)                                             CPPTESK_CALL_LOOP(method_name)
    #define CALL_LOOP_EX(full_method_name, iface, mode)                        CPPTESK_CALL_LOOP_EX(full_method_name, iface, mode)
#endif // HW_SHORT_NAMES

#define CPPTESK_CALL_STACK_OF(model) \
    ((model)._process_stack)

#define CPPTESK_CALL_STACK() \
    CPPTESK_CALL_STACK_OF(*this)

#ifdef HW_SHORT_NAMES
    #define CALL_STACK_OF(model) CPPTESK_CALL_STACK_OF(model)
    #define CALL_STACK()         CPPTESK_CALL_STACK()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Common part of model debug messages
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_COMMON_START_INFO(start_time) \
    "(iface=" << CPPTESK_GET_IFACE() << " cycle=" << std::dec << start_time << " pid=" << CPPTESK_GET_PROCESS().pid << ")"

#define __CPPTESK_COMMON_STOP_INFO(start_time, stop_time) \
    "(iface=" << CPPTESK_GET_IFACE() << " cycle=[" << std::dec << start_time << "," << stop_time << "] pid=" << CPPTESK_GET_PROCESS().pid << ")"

////////////////////////////////////////////////////////////////////////////////
// Start and Stop
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_START_LABEL $CPPTESK_START_PROCESS
#define __CPPTESK_START_VAR   _CPPTESK_START_PROCESS

#define __CPPTESK_CHECK_START() (void)__CPPTESK_START_VAR

#define __CPPTESK_INSIDE_PROCESS(process_stack) \
    /* Inform macroses that they are inside a process */ \
    bool  __cpptesk_inside_process = true; \
    void *__cpptesk_process_stack  = &(process_stack)

/** @def CPPTESK_START
 * Starts a model process.
 */
#define CPPTESK_START() \
    /* Inform macroses that they are inside a process */ \
    __CPPTESK_INSIDE_PROCESS(_process_stack); \
    \
    /* To check that CPPTESK_START() preceedes any process call */ \
    int __CPPTESK_START_VAR; \
    \
    /* Only one CPPTESK_START() is allowed within a specification function */ \
    __CPPTESK_START_LABEL: \
    \
    do { \
         /* To avoid the 'unused variable' warning */ \
         __CPPTESK_CHECK_START(); \
         /* To avoid the 'unused label' warnign */ \
         if(false) { goto __CPPTESK_START_LABEL; } \
         \
         CPPTESK_GET_PROCESS().start_call(); \
         CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_MORE, (CPPTESK_GET_PROCESS().call_stack_size() > 1 ? "[SEQUENTIAL:START]" : "[PARALLEL:START]") << \
             __CPPTESK_COMMON_START_INFO(time()) << ": " << CPPTESK_GET_PROCESS()); \
    } while(false)

/** @def CPPTESK_START_PROCESS
 * Starts a process.
 */
#define CPPTESK_START_PROCESS() \
    CPPTESK_START()

/** @def CPPTESK_START_STIMULUS
 * Starts a stimulus and launches the input interface adapter.
 */
#define CPPTESK_START_STIMULUS(call_mode) \
    CPPTESK_START(); \
    CPPTESK_RECV_STIMULUS(call_mode, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE())

/** @def CPPTESK_START_REACTION
 * Starts a reaction and launches the output interface adapter.
 */
#define CPPTESK_START_REACTION(call_mode) \
    CPPTESK_START(); \
    CPPTESK_SEND_REACTION(call_mode, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE())

/** @def CPPTESK_STOP
 * Stops a model process.
 */
#define CPPTESK_STOP() \
    do { \
        /* To check presense of CPPTESK_START() */ \
        __CPPTESK_CHECK_START(); \
         CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_MORE, (CPPTESK_GET_PROCESS().call_stack_size() > 1 ? "[SEQUENTIAL:STOP]" : "[PARALLEL:STOP]") << \
             __CPPTESK_COMMON_START_INFO(time()) << ": " << CPPTESK_GET_PROCESS()); \
        __CPPTESK_RETURN(::cpptesk::hw::COMPLETED); \
    } while(false)

/** @def CPPTESK_STOP_PROCESS
 * Stops a process.
 */
#define CPPTESK_STOP_PROCESS()  CPPTESK_STOP()

/** @def CPPTESK_STOP_STIMULUS
 * Stops a stimulus.
 */
#define CPPTESK_STOP_STIMULUS() CPPTESK_STOP()

/** @def CPPTESK_STOP_REACTION
 * Stops a reaction.
 */
#define CPPTESK_STOP_REACTION() CPPTESK_STOP()

#ifdef HW_SHORT_NAMES
    #define START()                   CPPTESK_START()
    #define START_PROCESS()           CPPTESK_START_PROCESS()
    #define START_STIMULUS(call_mode) CPPTESK_START_STIMULUS(call_mode)
    #define START_REACTION(call_mode) CPPTESK_START_REACTION(call_mode)
    #define STOP()                    CPPTESK_STOP()
    #define STOP_PROCESS()            CPPTESK_STOP_PROCESS()
    #define STOP_STIMULUS()           CPPTESK_STOP_STIMULUS()
    #define STOP_REACTION()           CPPTESK_STOP_REACTION()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Cycle, Delay, and Guard
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DELAY
 * Emulates a clock cycle of a unit.
 */
#define CPPTESK_CYCLE_OF(unit) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE, "CYCLE()"); \
        (unit).suspend(&CPPTESK_GET_PROCESS()); \
    } while(false)

/** @def CPPTESK_DELAY
 * Emulates a clock cycle.
 */
#define CPPTESK_CYCLE() \
    CPPTESK_CYCLE_OF(*this)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DELAY_OF
 * Emulates several clock cycles of a unit.
 */
#define CPPTESK_DELAY_OF(unit, delay) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE, "CYCLE()"); \
        while(CPPTESK_GET_PROCESS().point.timer < (::cpptesk::hw::cycle_t)(delay)) \
            { (unit).suspend(&CPPTESK_GET_PROCESS()); } \
    } while(false)

/** @def CPPTESK_DELAY
 * Emulates several clock cycles.
 */
#define CPPTESK_DELAY(delay) \
    CPPTESK_DELAY_OF(*this, delay)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GUARD_OF
 * Waits unit a condition becomes true.
 */
#define CPPTESK_GUARD_OF(unit, condition) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::GUARD, "WAIT(" #condition ")"); \
        while(!(condition)) \
            { (unit).suspend(&CPPTESK_GET_PROCESS()); } \
        /* Do not wait anymore */ \
        CPPTESK_GET_PROCESS().point.type = ::cpptesk::hw::CYCLE; \
        CPPTESK_GET_PROCESS().point.name = "CYCLE()"; \
    } while(false)

/** @def CPPTESK_GUARD
 * Waits until a condition becomes true.
 */
#define CPPTESK_GUARD(condition) \
    CPPTESK_GUARD_OF(*this, condition)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GUARD_TIMEOUT_OF
 * Waits until a condition becomes true or time is over.
 */
#define CPPTESK_GUARD_TIMEOUT_OF(unit, condition, limit) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::GUARD, "WAIT(" #condition ")"); \
        while(!(condition)) \
        { \
            /* Timeout violation */ \
            if(CPPTESK_GET_PROCESS().point.timer >= (::cpptesk::hw::cycle_t)(limit)) \
            { \
                CPPTESK_GET_PROCESS().timeout = true; \
                onTimeoutViolation(CPPTESK_GET_PROCESS()); \
                break; \
            } \
            (unit).suspend(&CPPTESK_GET_PROCESS()); \
        } \
        /* Do not wait anymore */ \
        CPPTESK_GET_PROCESS().point.type = ::cpptesk::hw::CYCLE; \
	CPPTESK_GET_PROCESS().point.name = "CYCLE()"; \
    } while(false)

/** @def CPPTESK_GUARD_TIMEOUT
 * Waits until a condition becomes true or time is over.
 */
#define CPPTESK_GUARD_TIMEOUT(condition, limit) \
    CPPTESK_GUARD_TIMEOUT_OF(*this, condition, limit)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_WAIT_OF
 * Alias of CPPTESK_GUARD_OF
 */
#define CPPTESK_WAIT_OF(unit, condition) \
    CPPTESK_GUARD_OF(unit, condition)

/** @def CPPTESK_WAIT
 * Alias of CPPTESK_GUARD
 */
#define CPPTESK_WAIT(condition) \
    CPPTESK_GUARD(condition)

/** @def CPPTESK_WAIT_TIMEOUT_OF
 * Alias of CPPTESK_GUARD_TIMEOUT_OF
 */
#define CPPTESK_WAIT_TIMEOUT_OF(unit, condition, limit) \
    CPPTESK_GUARD_TIMEOUT_OF(unit, condition, limit)

/** @def CPPTESK_WAIT_TIMEOUT
 * Alias of CPPTESK_GUARD_TIMEOUT
 */
#define CPPTESK_WAIT_TIMEOUT(condition, limit) \
    CPPTESK_GUARD_TIMEOUT(condition, limit)

////////////////////////////////////////////////////////////////////////////////

#ifdef HW_SHORT_NAMES
    #define CYCLE_OF(unit)                           CPPTESK_CYCLE_OF(unit)
    #define CYCLE()                                  CPPTESK_CYCLE()
    #define DELAY_OF(unit, delay)                    CPPTESK_DELAY_OF(unit, delay)
    #define DELAY(delay)                             CPPTESK_DELAY(delay)
    #define GUARD_OF(unit, condition)                CPPTESK_GUARD_OF(unit, condition)
    #define GUARD(condition)                         CPPTESK_GUARD(condition)
    #define GUARD_TIMEOUT_OF(unit, condition, limit) CPPTESK_GUARD_TIMEOUT_OF(unit, condition, limit)
    #define GUARD_TIMEOUT(condition, limit)          CPPTESK_GUARD_TIMEOUT(condition, limit)
    #define WAIT_OF(unit, condition)                 CPPTESK_WAIT_OF(unit, condition)
    #define WAIT(condition)                          CPPTESK_WAIT(condition)
    #define WAIT_TIMEOUT_OF(unit, condition, limit)  CPPTESK_WAIT_TIMEOUT_OF(unit, condition, limit)
    #define WAIT_TIMEOUT(condition, limit)           CPPTESK_WAIT_TIMEOUT(condition, limit)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Waiting Process
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_WAIT_PROCESS_STARTED(pid) \
    CPPTESK_WAIT(isFinished(pid) || isStarted(pid))

#define CPPTESK_WAIT_PROCESS_STARTED_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid) || isStarted(pid), limit)

#define CPPTESK_WAIT_PROCESS_WAITING(pid) \
    CPPTESK_WAIT(isFinished(pid) || isWaiting(pid))

#define CPPTESK_WAIT_PROCESS_WAITING_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid) || isWaiting(pid), limit)

#define CPPTESK_WAIT_PROCESS_ACTIVE(pid) \
    CPPTESK_WAIT(isFinished(pid) || isActive(pid))

#define CPPTESK_WAIT_PROCESS_ACTIVE_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid) || isActive(pid, limit))

#define CPPTESK_WAIT_PROCESS_FINISHED(pid) \
    CPPTESK_WAIT(isFinished(pid))

#define CPPTESK_WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid), limit)

#ifdef HW_SHORT_NAMES
    #define WAIT_PROCESS_STARTED(pid)                 CPPTESK_WAIT_PROCESS_STARTED(pid)
    #define WAIT_PROCESS_STARTED_TIMEOUT(pid, limit)  CPPTESK_WAIT_PROCESS_STARTED_TIMEOUT(pid, limit)
    #define WAIT_PROCESS_WAITING(pid)                 CPPTESK_WAIT_PROCESS_WAITING(pid)
    #define WAIT_PROCESS_WAITING_TIMEOUT(pid, limit)  CPPTESK_WAIT_PROCESS_WAITING_TIMEOUT(pid, limit)
    #define WAIT_PROCESS_ACTIVE(pid)                  CPPTESK_WAIT_PROCESS_ACTIVE(pid)
    #define WAIT_PROCESS_ACTIVE_TIMEOUT(pid, limit)   CPPTESK_WAIT_PROCESS_ACTIVE_TIMEOUT(pid, limit)
    #define WAIT_PROCESS_FINISHED(pid)                CPPTESK_WAIT_PROCESS_FINISHED(pid)
    #define WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit) CPPTESK_WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Process Attributes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GET_PRIORITY
 * Gets a process priority.
 */
#define CPPTESK_GET_PRIORITY() \
    ((unsigned)(CPPTESK_GET_PROCESS().getPriority()))

/** @def CPPTESK_SET_PRIORITY
 * Sets a process priority.
 */
#define CPPTESK_SET_PRIORITY(priority) \
    do { \
        CPPTESK_GET_PROCESS().setPriority((::cpptesk::hw::priority_t)priority); \
    } while(false)

/** @def CPPTESK_GET_INPUT_IFACE
 * Gets a process input interface.
 */
#define CPPTESK_GET_INPUT_IFACE() \
    (CPPTESK_GET_PROCESS().input_iface)

/** @def CPPTESK_SET_INPUT_IFACE
 * Sets a process input interface.
 */
#define CPPTESK_SET_INPUT_IFACE(iface) \
    do { \
        CPPTESK_GET_PROCESS().input_iface = &iface; \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define GET_PRIORITY()         CPPTESK_GET_PRIORITY()
    #define SET_PRIORITY(priority) CPPTESK_SET_PRIORITY(priority)
    #define GET_INPUT_IFACE()      CPPTESK_GET_INPUT_IFACE()
    #define SET_INPUT_IFACE(iface) CPPTESK_SET_INPUT_IFACE(iface)
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Model Events
///////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_RAISE_EVENT
 * Raises an event.
 */
#define CPPTESK_RAISE_EVENT(id) \
    raise(sequence::ModelEvent(id, _model_event_manager))

/** @def CPPTESK_CHECK_EVENT
 * Checks whether an event occures.
 */
#define CPPTESK_CHECK_EVENT(id) \
    check(sequence::ModelEvent(id, _model_event_manager))

/** @def CPPTESK_WAIT_EVENT
 * Waits until an event occures.
 */
#define CPPTESK_WAIT_EVENT(id) \
    CPPTESK_WAIT(CPPTESK_CHECK_EVENT(id))

/** @def CPPTESK_WAIT_EVENT_TIMEOUT
 * Waits until an event occures but not longer than an amount of time specified.
 */
#define CPPTESK_WAIT_EVENT_TIMEOUT(id, timeout) \
    CPPTESK_WAIT_TIMEOUT(CPPTESK_CHECK_EVENT(id), timeout)

#ifdef HW_SHORT_NAMES
    #define RAISE_EVENT(id)                 CPPTESK_RAISE_EVENT(id)
    #define CHECK_EVENT(id)                 CPPTESK_CHECK_EVENT(id)
    #define WAIT_EVENT(id)                  CPPTESK_WAIT_EVENT(id)
    #define WAIT_EVENT_TIMEOUT(id, timeout) CPPTESK_WAIT_EVENT_TIMEOUT(id, timeout)
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// The Basic Model Class
///////////////////////////////////////////////////////////////////////////////

namespace cpptesk {
namespace hw {

template<class UserModel /* Subclass */> class Model: public Synchronizer
{
public:
    typedef UserModel model_t;

    const static int DEFAULT_REACTION_TIMEOUT = 10;

    typedef void (UserModel::*process_t)(Process &process, Interface &iface, Message &message);

    typedef process_t stimulus_t;
    typedef process_t reaction_t;

    enum call_mode_t
    {
        PARALLEL,
        SEQUENTIAL
    };

    enum start_mode_t
    {
        /// Calling start() adds a new process into the pipeline; the process will be executed when cycle() is called.
        START_NORMAL        = 0x0000 | NULL_PRIORITY,
        /// Calling start() adds a new process into the pipeline and executes the first stage of the process within the current cycle.
        START_IMMEDIATELY   = 0x0100 | NULL_PRIORITY,
        /// Request interface when starting a process.
        START_REQUEST_IFACE = 0x0200 | NULL_PRIORITY,
    };

    Model();
    virtual ~Model();

    void registerComponent(const Component &component);

    void registerInterfaceAdapter(Interface &iface, process_t adapter, const std::string &name);

    inline void registerInputAdapter(Interface &input, process_t serializer, const std::string &name)
        { registerInterfaceAdapter(input, serializer, name); }

    inline void registerOutputAdapter(Interface &output, process_t deserializer, const std::string &name)
        { registerInterfaceAdapter(output, deserializer, name); }

    void registerInterfaceAdapter(Interface &iface, model_t &model, process_t adapter, const std::string &name);

    inline void registerInputAdapter(Interface &input, model_t &model, process_t adapter, const std::string &name)
        { registerInterfaceAdapter(input, model, adapter, name); }

    inline void registerOutputAdapter(Interface &output, model_t &model, process_t adapter, const std::string &name)
        { registerInterfaceAdapter(output, model, adapter, name); }

public:

    inline Process::CallDescriptor get_call() const { return _call; }
    inline void prepare_call(const Process::CallDescriptor &call) { _call = call; }
    inline void reset_call() { _call.reset(); }

protected:

    /**
     * Descriptor of a current process call:
     *
     * CALL_PROCESS():  prepare_call(DESCRIPTOR);
     * start():         process.prepare_call(get_call()); reset_call();
     * START_PROCESS(): process.start_call();
     * STOP_PROCESS():  process.stop_call();
     */
    Process::CallDescriptor _call;

public:

///////////////////////////////////////////////////////////////////////////////
// Process creation and destruction methods
///////////////////////////////////////////////////////////////////////////////

    /**
     * Starts a new process for a given stimulus.
     */
    pid_t start(stimulus_t stimulus, Interface &iface, Message &message, int mode = START_NORMAL);

    /**
     * Starts a new process for a given stimulus with the empty message.
     */
    pid_t start(stimulus_t stimulus, Interface &iface, int mode = START_NORMAL);

    /**
     * Processes a stimulus within a given process. Calling start(&UserModel::stimulus, process, message)
     * is fully equivalent to userModel.stimulus(process, iface, message).
     */
    pid_t start(stimulus_t stimulus, Process &process, Interface &iface, Message &message = CPPTESK_DEFAULT_MESSAGE);

    /**
     * Stops a process.
     */
    void stop(pid_t pid);

    void reset();

    size_t size() const;

///////////////////////////////////////////////////////////////////////////////
// Methods for checking process status
///////////////////////////////////////////////////////////////////////////////

    bool isStarted(pid_t pid)  const;
    bool isWaiting(pid_t pid)  const;
    bool isActive(pid_t pid)   const;
    bool isFinished(pid_t pid) const;

///////////////////////////////////////////////////////////////////////////////
// System clock modelling
///////////////////////////////////////////////////////////////////////////////

    cycle_t time() const;

    void cycle();
    void delay(cycle_t n);

///////////////////////////////////////////////////////////////////////////////
// History of Stimuli & Reactions
///////////////////////////////////////////////////////////////////////////////

public:

    /**
     * Register a stimulus or a reaction.
     */
    void registerStimulus(const Stimulus &stimulus);
    void registerReaction(const Reaction &reaction);

    void registerStimulus(const Interface &iface);
    void registerReaction(error_type_t error, const Interface &iface, const std::string &info, unsigned type = -1u);

    void clearReactions();

    /**
     * Returns the first error (if there is any) having occured starting from the previous call of verdict().
     * Do not clear the set of reactions.
     */
    std::string error(error_mode_t mode = LAST_REACTION) const;

    /**
     * Returns a verdict for a set of reactions having occured starting from the previous call. 
     * Clears the set of reactions to be analyzed.
     */
    bool verdict();

    /**
     * Returns a list of reactions.
     * @depricated
     */
    const std::vector<Reaction>& getReactions() const;

protected:

    EventHistory _history;

////////////////////////////////////////////////////////////////////////////////
// Model Events Management (for Temporal Sequence Recognition)
////////////////////////////////////////////////////////////////////////////////

public:

    void raise(const sequence::ModelEvent &event) { _model_event_manager.add(event); }
    bool check(const sequence::ModelEvent &event) const { return _model_event_manager.contains(event); }

protected:

    sequence::ModelEventManager _model_event_manager;

////////////////////////////////////////////////////////////////////////////////
// Working with input and output interfaces
////////////////////////////////////////////////////////////////////////////////

protected:

    /// Pseudo interface for reactions that are not assosiated with existing interfaces.
    CPPTESK_DECLARE_INOUT(CPPTESK_DEFAULT_IFACE);

    /// Processes a stimulus: serializes a message and applies it to an input interface.
    CPPTESK_DECLARE_PROCESS(onStimulus);
    /// Processes a reaction: detects a reaction, deserializes a message from an output interface and calls a comparator.
    CPPTESK_DECLARE_PROCESS(onReaction);

    void setReactionTimeout(unsigned timeout);

    unsigned getReactionTimeout() const;

////////////////////////////////////////////////////////////////////////////////
// Callbacks
////////////////////////////////////////////////////////////////////////////////

    // Process callbacks
    virtual void onPreconditionViolation(Process &process);
    virtual void onPostconditionViolation(Process &process);
    virtual void onTimeoutViolation(Process &process);
    virtual void onNormalCompletion(Process &process);
    virtual void onEveryCycle(Process &process);

    // Model callbacks
    virtual void onEveryCycle();

    /**
     * Checks data integrity of a model. The callback is overloaded by a model.
     */
    virtual bool invariant();

////////////////////////////////////////////////////////////////////////////////
// Synchronization methods
////////////////////////////////////////////////////////////////////////////////

public:

    /**
     * Initializes simulation.
     */
    virtual void initialize();

    /**
     * Finalizes simulation.
     */
    virtual void finalize();

    /**
     * Inputs synchronizer: sets input signals of a design.
     */
    virtual void setInputs();

    /**
     * Outputs synchronizer: gets output signals of a design.
     */
    virtual void getOutputs();

    /**
     * Time synchronizer: synchronizes execution threads of a model and simulator.
     */
    virtual void simulate();

    /**
     * Registers synchronizer.
     */
    void setSynchronizer(Synchronizer *synchronizer);

private:
////////////////////////////////////////////////////////////////////////////////
// Auxiliary methods
////////////////////////////////////////////////////////////////////////////////
    static void launch(UserModel *model, stimulus_t stimulus, Process *process, Interface *iface, Message *message);

// IMPORTANT: Should be public (CYCLE_EX)
public:
    Process* fork(Interface &iface, Message &message);
    void resume(Process *process);
    void suspend(Process *process);

protected:
////////////////////////////////////////////////////////////////////////////////
// Data
////////////////////////////////////////////////////////////////////////////////
    typedef std::vector<Component*> components_list_t;
    components_list_t _components;

    class ProcessWrapper
    {
    public:
        ProcessWrapper(model_t &model, process_t process, const std::string name):
            _model(model), _process(process), _name(name) {}

        CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(call)
        {
            // Since it is an internal call, it is transparent to an user
            _model._call = process.call_stack_top(); _model._call.name = _name;
            _model.start(_process, CPPTESK_GET_PARAMETERS());
        }
    private:
        model_t &_model;
        process_t _process;
        std::string _name;
    };

    typedef std::map<Interface*, ProcessWrapper*> adapters_map_t;
    adapters_map_t _adapters;

    cycle_t _reaction_timeout;

    pid_t _new_pid;

    cycle_t _clock;

    typedef std::map<pid_t, Process*> process_table_t;
    process_table_t _process_table;

    Process *_main_process;

public:

    /// The main loop of a test system.
    Process &main() { return *_main_process; }

public:
    /// IMPORTANT: The process stack should be public.
    process_stack_t _process_stack;

    // TODO:
    std::vector<Process::PointDescriptor> getPipelineSlice() const
    {
        std::vector<Process::PointDescriptor> slice(_process_table.size());

        for(process_table_t::const_iterator i = _process_table.begin(); i != _process_table.end(); i++)
            { slice.push_back(i->second->point);  }

	return slice;
    }

////////////////////////////////////////////////////////////////////////////////
// Simulator Synchronizer
////////////////////////////////////////////////////////////////////////////////
private:

    Synchronizer *_synchronizer;
};

template<class UserModel>
Model<UserModel>::Model():
    CPPTESK_DEFAULT_IFACE("<default>"),
    _reaction_timeout(DEFAULT_REACTION_TIMEOUT),
    _new_pid(0),
    _clock(0),
    _synchronizer(0)
{
    // The main process of the model
    _main_process = new Process(-1, CPPTESK_DEFAULT_IFACE, CPPTESK_DEFAULT_MESSAGE);
    CPPTESK_ASSERTION(_main_process, "No memory to create a process");

    // Pseudo call of the main process.
    _main_process->prepare_call(Process::CallDescriptor("<cpptesk_main>"));
    _main_process->start_call();

    // It is a bottom of the process stack
    _process_stack.push_back(_main_process);
}

template<class UserModel>
Model<UserModel>::~Model()
{
    // Reset the model's pipeline
    reset();

    // Delete the interface adapters
    for(typename adapters_map_t::iterator i = _adapters.begin(); i != _adapters.end(); i++)
        { delete i->second; }
    _adapters.clear();

    // Delete the synchronizer
    if(_synchronizer)
        { delete _synchronizer; }

    // Delete the main process
    if(_main_process)
    {
        // Finish pseudo call of the main process
        _main_process->stop_call();
        // Delete the main process
        delete _main_process;
    }
}

template<class UserModel>
void Model<UserModel>::registerComponent(const Component &component)
{
    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "Register component " << component.getName());
    _components.push_back(const_cast<Component*>(&component));
}

template<class UserModel>
void Model<UserModel>::registerInterfaceAdapter(Interface &iface, process_t adapter, const std::string &name)
{
    registerInterfaceAdapter(iface, dynamic_cast<model_t&>(*this), adapter, name);
}

template<class UserModel>
void Model<UserModel>::registerInterfaceAdapter(Interface &iface, model_t &model, process_t adapter, const std::string &name)
{
    typename adapters_map_t::iterator i = _adapters.find(&iface);

    // Deallocate the previously registered adapter
    if(i != _adapters.end() && !(i->second))
        { delete i->second; }	

    // Register the new one
    _adapters[&iface] = new ProcessWrapper(model, adapter, name);
}

template<class UserModel>
void Model<UserModel>::launch(UserModel *model, stimulus_t stimulus, Process *process, Interface *iface, Message *message)
{
    CPPTESK_DEBUG_PRINTF_FILE_LINE(DEBUG_TOOL, "Model::launch: model=%p, process=%p, iface=%p, message=%p", model, process, iface, message);

    CPPTESK_ASSERTION(process != NULL, "Parameter process is null");
    CPPTESK_ASSERTION(iface   != NULL, "Parameter iface is null");
    CPPTESK_ASSERTION(message != NULL, "Parameter message is null");

    (model->*stimulus)(*process, *iface, *message);
}

template<class UserModel>
pid_t Model<UserModel>::start(stimulus_t stimulus, Interface &iface, Message &message, int mode)
{
    Process *process = fork(iface, message);
    CPPTESK_ASSERTION(process != NULL, "Process has not forked");

    // Set a call descriptor
    process->prepare_call(get_call()); reset_call();

    priority_t priority = (priority_t)(mode & __CPPTESK_PRIORITY_MASK); 
    process->setPriority(priority == NULL_PRIORITY ? NORMAL_PRIORITY : priority);

    mode &= ~__CPPTESK_PRIORITY_MASK;

    if((mode & START_REQUEST_IFACE) != 0 && iface.isInput())
        { iface.request(*process, *(process->input_message)); }

    Interface *input_iface = process->input_iface;   // The same interface
    Message *input_message = process->input_message; // The message's clone

    // If a process finishes unexpectedly, the execution switches to the main loop
    process->ucp->uc_link = _main_process->ucp;
    makecontext(process->ucp, (void(*)(void))Model<UserModel>::launch, 6, this, stimulus, process, input_iface, input_message);

    // Execute the first stage of the process within the current cycle if START_IMMEDIATELY is set
    if((mode & START_IMMEDIATELY) != 0)
    {
        Process *this_process = _process_stack.back();

        resume(process);

        // There is no need to check the execution flow consistency, because
        // if a process finishes unexpectedly, the code below is not executed.
        CPPTESK_ASSERTION(_process_stack.back() == this_process, "Process " << process <<
            " has unexpectedly finished (probably, STOP_PROCESS() or STOP_STIMULUS() are missing)");
    }

    // Return PID of the newly created process
    return process->pid;
}

template<class UserModel>
pid_t Model<UserModel>::start(stimulus_t stimulus, Interface &iface, int mode)
{
    return start(stimulus, iface, CPPTESK_DEFAULT_MESSAGE, mode);
}

template<class UserModel>
pid_t Model<UserModel>::start(stimulus_t stimulus, Process &process, Interface &iface, Message &message)
{
    // Set a call descriptor
    process.prepare_call(get_call()); reset_call();

    ((UserModel*)this->*stimulus)(process, iface, message);

    return process.pid;
}

template<class UserModel>
void Model<UserModel>::stop(pid_t pid)
{
    process_table_t::iterator i = _process_table.find(pid);

    if(i != _process_table.end())
    {
        delete i->second;
        _process_table.erase(i);
    }
}

template<class UserModel>
bool Model<UserModel>::isStarted(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);
    CPPTESK_ASSERTION(i != _process_table.end(), "Process has not been found");

    return i->second->isStarted();
}

template<class UserModel>
bool Model<UserModel>::isWaiting(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);
    CPPTESK_ASSERTION(i != _process_table.end(), "Process has not been found");

    return i->second->isWaiting();
}

template<class UserModel>
bool Model<UserModel>::isActive(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);
    CPPTESK_ASSERTION(i != _process_table.end(), "Process has not been found");

    return i->second->isActive();
}

template<class UserModel>
bool Model<UserModel>::isFinished(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);

    return i == _process_table.end();
}

template<class UserModel>
cycle_t Model<UserModel>::time() const
{
    return _clock;
}

template<class UserModel>
size_t Model<UserModel>::size() const
{
    return _process_table.size();
}

template<class UserModel>
void Model<UserModel>::cycle()
{
    std::vector<pid_t> to_kill;
    std::vector<Process*> processes;

    // Callback
    onEveryCycle();

    // Check a model's invariant
    CPPTESK_ASSERTION(invariant(), "Model invariant violation");

    // Process the model event manager (reset the set of events)
    _model_event_manager.cycle();

    // Process components
    for(components_list_t::iterator i = _components.begin(); i != _components.end(); i++)
        { (*i)->cycle(); }

    // List of processes
    for(process_table_t::iterator i = _process_table.begin(); i != _process_table.end(); i++)
        { processes.push_back(i->second); }

    // Sort processes in order of decreasing priorities
    stable_sort(processes.begin(), processes.end(), greaterProcessPriority);

    for(std::vector<Process*>::iterator i = processes.begin(); i != processes.end(); i++)
    {
        Process *process = *i;

        // The process status after starting is usually EXECUTING, but there are some exceptions
        if(process->status == EXECUTING)
        {
            // Execute a single stage of the process
            resume(process);
            // For dumping the process stack if the assertion below is violated
            __CPPTESK_INSIDE_PROCESS(_process_stack);
            // Checking execution flow consistency
            CPPTESK_ASSERTION(_process_stack.back() == _main_process, "Process " << *process <<
                " has unexpectedly finished (probably, STOP_PROCESS() or STOP_STIMULUS() are missing)");
        }

        if(process->status != EXECUTING)
            { to_kill.push_back(process->pid); }
    }

    for(std::vector<pid_t>::iterator j = to_kill.begin(); j != to_kill.end(); j++)
        { stop(*j); }

    // Set input signals
    setInputs();

    // Simulate a cycle
    simulate(); _clock++;

    // Get output signals
    getOutputs();
}

template<class UserModel>
void Model<UserModel>::reset()
{
    for(process_table_t::iterator i = _process_table.begin(); i != _process_table.end(); i++)
        { delete i->second; }

    _process_table.clear();
}

template<class UserModel>
void Model<UserModel>::delay(cycle_t n)
{
    for(cycle_t i = 0; i < n; i++)
        { cycle(); }
}

template<class UserModel>
const std::vector<Reaction>& Model<UserModel>::getReactions() const
{
    return _history.getReactions();
}

////////////////////////////////////////////////////////////////////////////////
// Working with input and output interfaces
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
CPPTESK_DEFINE_PROCESS(Model<UserModel>::onStimulus)
{
    CPPTESK_START_PROCESS();

    time_t start_time = time();

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[STIMULUS:START]" << __CPPTESK_COMMON_START_INFO(start_time) << ": " << CPPTESK_GET_MESSAGE());

    // If a model does not have interface adapters, then do nothing
    if(_adapters.empty())
    {
        // Do nothing
    }
    else
    {
        // Otherwise, try to find an appropriate one
        typename adapters_map_t::iterator i = _adapters.find(&iface);

        // Ensure that an interface adapter has been found
        CPPTESK_ASSERTION(i != _adapters.end(), "Adapter has not been found");
        CPPTESK_ASSERTION(i->second != 0, "Adapter is null");

        // Call the serializer having been found
        (i->second->call)(CPPTESK_GET_PARAMETERS());
    }

    time_t stop_time = time();

    Stimulus stimulus(CPPTESK_GET_IFACE(), start_time, stop_time);
    registerStimulus(stimulus);

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[STIMULUS:STOP]" << __CPPTESK_COMMON_STOP_INFO(start_time, stop_time));

    CPPTESK_STOP_PROCESS();
}

template<class UserModel>
CPPTESK_DEFINE_PROCESS(Model<UserModel>::onReaction)
{
    CPPTESK_START_PROCESS();

    time_t start_time = time();
    time_t stop_time  = time();

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[REACTION:START]" << __CPPTESK_COMMON_START_INFO(start_time) << ": " << CPPTESK_GET_MESSAGE());

    // Emulate the data transmission
    CPPTESK_SET_PRIORITY(CPPTESK_GET_MESSAGE().getPriority());

    if((CPPTESK_GET_MESSAGE().getInterlock() & LOCK_HOLD)  != 0)
        { CPPTESK_WAIT_PROCESS_FINISHED(CPPTESK_GET_MESSAGE().getPID()); }
    if((CPPTESK_GET_MESSAGE().getInterlock() & LOCK_DELAY) != 0)
        { CPPTESK_DELAY(CPPTESK_GET_MESSAGE().getDelay()); }

    // If a model does not have interface adapters,
    // then register a normal (correct) reaction
    if(_adapters.empty())
    {
        Reaction reaction(NORMAL_REACTION, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE(), start_time, stop_time, "");

        CPPTESK_DEBUG_PRINT(DEBUG_INFO, reaction);
        registerReaction(reaction);
    }
    else
    {
#ifndef CPPTESK_CLONE_INPUT_MESSAGE
        // The default handler requies clonning of input messages
        CPPTESK_ASSERTION(false, "Default reaction handler does not support non-cloning messages");
#endif
        // Otherwise (there are interface adapters), try to find an appropriate one
        typename adapters_map_t::iterator i = _adapters.find(&iface);

        // Ensure that an interface adapter has been found
        CPPTESK_ASSERTION(i != _adapters.end(), "Adapter has not been found");
        CPPTESK_ASSERTION(i->second != 0, "Adapter is null");

        Message &spec_msg =  (CPPTESK_GET_MESSAGE());
        Message &impl_msg = *(CPPTESK_GET_MESSAGE().clone());

        // Invalidate the message's fields to be able to check the deserializer
        impl_msg.invalidateFields();
        // Call the deserializer having been found
        (i->second->call)(CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), impl_msg /* input/output */);

        std::stringstream buffer;
        error_type_t error = NORMAL_REACTION;

        if(process.timeout)
        {
            error = MISSING_REACTION;
        }
        else
        {
            std::string info = spec_msg.compare(impl_msg);

            if(!info.empty())
            {
                error = INCORRECT_REACTION;
                buffer << info << " ";
            }
        }

        if(error != NORMAL_REACTION)
            { buffer << spec_msg.getExplanation(); }

        stop_time = time();

        // Register the reaction
        Reaction reaction(error, CPPTESK_GET_IFACE(), spec_msg, impl_msg, start_time, stop_time, buffer.str());

        CPPTESK_DEBUG_PRINT((error != NORMAL_REACTION ? DEBUG_ERROR : DEBUG_INFO), reaction);
        registerReaction(reaction);

        // Delete the clone
        delete &impl_msg;
    }

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[REACTION:STOP]" << __CPPTESK_COMMON_STOP_INFO(start_time, stop_time));

    CPPTESK_STOP_PROCESS();
}

////////////////////////////////////////////////////////////////////////////////
// Stimuli & Reactions
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
void Model<UserModel>::registerStimulus(const Stimulus &stimulus)
{
    _history.registerStimulus(stimulus);
}

template<class UserModel>
void Model<UserModel>::registerReaction(const Reaction &reaction)
{
    _history.registerReaction(reaction);
}

template<class UserModel>
void Model<UserModel>::registerStimulus(const Interface &iface)
{
    Stimulus stimulus(iface, time(), time());
    registerStimulus(stimulus);
}

template<class UserModel>
void Model<UserModel>::registerReaction(error_type_t error, const Interface &iface, const std::string &info, unsigned type)
{
    Reaction reaction(error, iface, time(), time(), info, type);
    registerReaction(reaction);
}

template <class UserModel>
void Model<UserModel>::clearReactions()
{
    _history.clearReactions();
}

template <class UserModel>
std::string Model<UserModel>::error(error_mode_t mode) const
{
    return _history.error(mode);
}

/**
 * Returns a verdict for a set of reactions having been occured starting from the previous call. 
 */
template <class UserModel>
bool Model<UserModel>::verdict()
{
    return _history.verdict();
}

template<class UserModel>
void Model<UserModel>::setReactionTimeout(unsigned timeout)
{
    _reaction_timeout = timeout;
}

template<class UserModel>
unsigned Model<UserModel>::getReactionTimeout() const
{
    return _reaction_timeout;
}

////////////////////////////////////////////////////////////////////////////////
// Callbacks
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
void Model<UserModel>::onPreconditionViolation(Process &process)
{
    std::stringstream buffer;
    buffer << "(" << process.point.file << ":" << process.point.line << ")";
    registerReaction(ASSERTION_VIOLATION, CPPTESK_DEFAULT_IFACE, buffer.str());
}

template<class UserModel>
void Model<UserModel>::onPostconditionViolation(Process &process)
{
    std::stringstream buffer;
    buffer << "(" << process.point.file << ":" << process.point.line << ")";
    registerReaction(ASSERTION_VIOLATION, CPPTESK_DEFAULT_IFACE, buffer.str());
}

template<class UserModel>
void Model<UserModel>::onTimeoutViolation(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
}

template<class UserModel>
void Model<UserModel>::onNormalCompletion(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
}

template<class UserModel>
void Model<UserModel>::onEveryCycle(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
}

template<class UserModel>
void Model<UserModel>::onEveryCycle()
{
}

template<class UserModel>
bool Model<UserModel>::invariant()
{
    return true;
}

////////////////////////////////////////////////////////////////////////////////
// Synchronization methods
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
void Model<UserModel>::initialize()
{
    if(_synchronizer) { _synchronizer->initialize(); }
}

template<class UserModel>
void Model<UserModel>::finalize()
{
    if(_synchronizer) { _synchronizer->finalize(); }

    _history.traceStatistics();
}

template<class UserModel>
void Model<UserModel>::setInputs()
{
    if(_synchronizer) { _synchronizer->setInputs(); }
}

template<class UserModel>
void Model<UserModel>::getOutputs()
{
    if(_synchronizer) { _synchronizer->getOutputs(); }
}

template<class UserModel>
void Model<UserModel>::simulate()
{
    if(_synchronizer) { _synchronizer->simulate(); }
}

template<class UserModel>
void Model<UserModel>::setSynchronizer(Synchronizer *synchronizer)
{
    _synchronizer = synchronizer;
}

////////////////////////////////////////////////////////////////////////////////
// Auxiliary methods
////////////////////////////////////////////////////////////////////////////////

template<class UserModel>
Process* Model<UserModel>::fork(Interface &iface, Message &message)
{
    // IMPORTANT: A newly created process should have a greater PID than the
    // previously launched ones. This guarantees that reset of a control signal
    // (strobe) is done before a new process set the signal.
    Process *process = new Process(_new_pid++, iface, message);
    CPPTESK_ASSERTION(process, "No memory");

    CPPTESK_DEBUG_PRINT_FILE_LINE(DEBUG_TOOL, "Model::fork: create " << *process);

    _process_table.insert(std::pair<pid_t, Process*>(process->pid, process));

    return process;
}

template<class UserModel>
void Model<UserModel>::resume(Process *process)
{
    Process *from;
    Process *jump;

    bool new_process = !process->isStarted();

    CPPTESK_ASSERTION(process, "Parameter process is null");

    // Update timers
    process->cycle();

    CPPTESK_ASSERTION(!_process_stack.empty(), "Execution flow inconsistency: empty process stack");

    from = _process_stack.back();
    _process_stack.push_back(process);
    jump = _process_stack.back();

    CPPTESK_DEBUG_PRINT_FILE_LINE(DEBUG_TOOL, "Model::resume: switch from " << *from << " to " << *jump);

    swapcontext(from->ucp, jump->ucp);

    if(new_process && process->status != EXECUTING)
    {
        CPPTESK_DEBUG_PRINT(DEBUG_WARN, "Process " << *process
            << " has finished within one cycle (probably, CYCLE() or DELAY() are missing)");
    }

    // Process callbacks
    switch(process->status)
    {
    case UNPREDICT: { onPreconditionViolation  (*process); break; }
    case VIOLATION: { onPostconditionViolation (*process); break; }
    case COMPLETED: { onNormalCompletion       (*process); break; }
    case EXECUTING: { onEveryCycle             (*process); break; }
    default:        { CPPTESK_ASSERTION(false, "Unknown process status"); }
    }
}

template<class UserModel>
void Model<UserModel>::suspend(Process *process)
{
    Process *from;
    Process *jump;

    CPPTESK_ASSERTION(process != NULL, "Parameter process is null");
    CPPTESK_ASSERTION(_process_stack.size() > 1, "Execution flow inconsistency");

    from = _process_stack.back();
    _process_stack.pop_back();
    jump = _process_stack.back();

    CPPTESK_DEBUG_PRINT_FILE_LINE(DEBUG_TOOL, "Model::suspend: switch from " << *from << " to " << *jump);

    CPPTESK_ASSERTION(from == process, "Execution flow inconsistency");

    swapcontext(from->ucp, jump->ucp);
}

CPPTESK_MODEL(EMPTY_MODEL) {};

}} // namespace cpptesk::hw
