/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Sequence recognizer
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>

#include <hw/sequence/delay.hpp>
#include <hw/sequence/repeat.hpp>
#include <hw/sequence/sequence.hpp>

#include <list>

namespace cpptesk {
namespace hw {
namespace sequence {

////////////////////////////////////////////////////////////////////////////////
// Sequence Recognizer
////////////////////////////////////////////////////////////////////////////////

class SequenceRecognizer
{
public:
    SequenceRecognizer(const Sequence &sequence):
        sequence(sequence),
	initial_delay(sequence.getDelay()),
	time_constraints(sequence.getTime()) { reset(); }

    SequenceRecognizer(const Sequence &sequence, const Delay &delay, const Delay &time):
        sequence(sequence),
	initial_delay(delay),
        time_constraints(time) { reset(); }

    virtual ~SequenceRecognizer() { reset(); }

    /// The main interface method. Returns true iff new match is found.
    bool cycle();

    bool isWorking()   const { return status == START; }
    bool isFinished()  const { return status == STOP;  }
    bool isSucceeded() const { return status == STOP && count >  0; }
    bool isFailed()    const { return status == STOP && count == 0; }

    /// Returns the overall number of matches.
    unsigned getMatchCount() const { return count; }
	
    /// Returns the overall number of cycles having been taken for the sequence acceptance/rejection.
    unsigned getTime() const { return time; }

protected:
    /// The main recognition method (should be overloaded in the subclasses). Returns true iff new match is found.
    virtual bool onCycle();

    /// Starts the recognition process (not used).
    void start() { status = START; }
    /// Stops the recognition process.
    void stop()  { status = STOP; }

    /// Increments the match count.
    bool match(bool value = true)
    {
        if(value) { count++; }
        return value;
    }

    /// Increments the recognition time.
    void tick()
    {
        time++;
        if(delay.min > 0) { delay.min--; }
        if(delay.max > 0) { delay.max--; }
    }

    /// Resets the recognizer's state (destroys the child recognizers, in particular).
    void reset();

protected:
    /// Data type for recognizer's status.
    enum status_t { STOP, START };

protected:
    /// The recognizer's status (START if it has been started; STOP otherwise).
    status_t status;
    /// The overall number of matches.
    unsigned count;
    /// The sequence being recognized.
    const Sequence &sequence;
   /// Initial delay.
    Delay initial_delay;
    /// Time constraints for recognition.
    Delay time_constraints;
    /// Delay tracker.
    Delay delay;
    /// Recognition time.
    unsigned time;
    /// Delay timeout.
    bool timeout;

    /// Child recognizers.
    std::list<SequenceRecognizer*> alt;
};

class EventRecognizer: public SequenceRecognizer
{
public:
    EventRecognizer(const Sequence &sequence):
        SequenceRecognizer(sequence) {}
    virtual ~EventRecognizer() {}

protected:
    virtual bool onCycle()
    {
        bool new_match = match(sequence.getEvent().match());

        stop();

        return new_match;
    }
};

class GuardRecognizer: public SequenceRecognizer
{
public:
    GuardRecognizer(const Sequence &sequence):
        SequenceRecognizer(sequence) {}
    virtual ~GuardRecognizer() {}

protected:
    virtual bool onCycle()
    {
        bool new_match = match(sequence.getEvent().match());

        if(new_match) { stop(); }

        return new_match;
    }
};

class SeqRecognizer: public SequenceRecognizer
{
public:
    SeqRecognizer(const Sequence &sequence):
        SequenceRecognizer(sequence),
        i(sequence.begin()),
	main(new SequenceRecognizer(sequence.getChild())) {}

    SeqRecognizer(const Sequence &sequence, const Sequence::const_iterator i):
        SequenceRecognizer(sequence),
	i(i),
	main(new SequenceRecognizer(*i)) {}

    virtual ~SeqRecognizer()
    {
        delete main;

	for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end(); i++)
            { delete *i; }
    }

protected:
    virtual bool onCycle()
    {
        bool new_match = false;
        bool main_inner_match = main->cycle();

	if(main_inner_match)
	{
	    if(main->isWorking())
	    {
	        if(++i != sequence.end())
		    { alt.push_back(new SeqRecognizer(sequence, i)); }

		i--;
	    }
	    else
            {
	        i++;

	        if(i == sequence.end())
		    { new_match = true; }
		else
		    { delete main; main = new SequenceRecognizer(*i); }
            }
	}

        for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end(); i++)
	{
	    new_match |= (*i)->cycle();

	    if((*i)->isFinished())
	        { delete *i; alt.erase(i--); }
	}

        if(main->isFinished() && alt.empty())
            { stop(); }

        return match(new_match);
    }

protected:
    Sequence::const_iterator i;
    SequenceRecognizer *main;
    std::list<SequenceRecognizer*> alt;
};

class AndRecognizer: public SequenceRecognizer
{
public:
    AndRecognizer(const Sequence &sequence):
        SequenceRecognizer(sequence),
        lhs(new SequenceRecognizer(sequence.getLhsChild())),
        rhs(new SequenceRecognizer(sequence.getRhsChild())) {}

    virtual ~AndRecognizer()
    {
        delete lhs;
	delete rhs;
    }

protected:
    virtual bool onCycle()
    {
        bool new_lhs_match = lhs->cycle();
	bool new_rhs_match = rhs->cycle();

        if(lhs->isFinished() && rhs->isFinished()) { stop(); }

	return match((new_lhs_match && rhs->getMatchCount() > 0)
	          || (new_rhs_match && lhs->getMatchCount() > 0));
    }

protected:
    SequenceRecognizer *lhs;
    SequenceRecognizer *rhs;
};

class OrRecognizer: public SequenceRecognizer
{
public:
    OrRecognizer(const Sequence &sequence):
        SequenceRecognizer(sequence),
        lhs(new SequenceRecognizer(sequence.getLhsChild())),
        rhs(new SequenceRecognizer(sequence.getRhsChild())) {}

    virtual ~OrRecognizer()
    {
        delete lhs;
        delete rhs;
    }

protected:
    virtual bool onCycle()
    {
        bool new_lhs_match = lhs->cycle();
        bool new_rhs_match = rhs->cycle();

	if(lhs->isFinished() && rhs->isFinished()) { stop(); }

        return match(new_lhs_match || new_rhs_match);
    }

protected:
    SequenceRecognizer *lhs;
    SequenceRecognizer *rhs;
};

class RepeatRecognizer: public SequenceRecognizer
{
public:
    RepeatRecognizer(const Sequence &sequence):
        SequenceRecognizer(sequence),
	repeat(sequence.getRepeat()),
        main(new SequenceRecognizer(sequence.getChild())) {}

    RepeatRecognizer(const Sequence &sequence, const Repeat &repeat):
        SequenceRecognizer(sequence),
        repeat(repeat),
	main(new SequenceRecognizer(sequence.getChild())) {}

    virtual ~RepeatRecognizer()
    {
        delete main;

        for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end(); i++)
	    { delete *i; }
    }

protected:
    virtual bool onCycle()
    {
        bool new_match = false;
        bool new_inner_match = main->cycle();

        if(new_inner_match)
        {
            if(main->isWorking())
	    {
	        if(repeat.max > 1)
		{
		    Repeat dec_repeat((repeat.min > 0 ? repeat.min - 1 : 0),
			              (repeat.max > 0 ? repeat.max - 1 : 0));

		    alt.push_back(new RepeatRecognizer(sequence, dec_repeat));
		}
	    }
	    else
	    {
	        if(repeat.min > 0) { repeat.min--; }
	        if(repeat.max > 0) { repeat.max--; }

                if(repeat.min == 0)
                    { new_match = true; }

		if(repeat.max > 0)
		    { delete main; main = new SequenceRecognizer(sequence.getChild()); }
	    }
        }

        for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end(); i++)
	{
            new_match |= (*i)->cycle();

	    if((*i)->isFinished())
	        { delete *i; alt.erase(i--); }
	}

        if(main->isFinished() && alt.empty())
	    { stop(); }

        return match(new_match);
    }

protected:
    Repeat repeat;
    SequenceRecognizer *main;
    std::list<SequenceRecognizer*> alt;
};

}}} // namespace cpptesk::hw::sequence

