/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Basic class for a test case
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/model.hpp>

#include <assert.h>
#include <stdint.h>

/** @def CPTESK_TEST
 * Macro declaring a test.
 */
#define CPPTESK_TEST(UserTest, UserModel) class UserTest: public ::cpptesk::hw::Test<UserModel>

/** @def CPPTESK_MODEL_UNDER_TEST
 * Macro for initialization of a model under test.
 */
#define CPPTESK_MODEL_UNDER_TEST(UserModel, UserModelInstance) ::cpptesk::hw::Test<UserModel>(UserModelInstance)

namespace cpptesk {
namespace hw {

/**
 * Test for a UserModel.
 */
template<class UserModel> class Test 
{
public:
    /**
     * Constructor.
     */
    Test(UserModel &model_under_test);

    /**
     * Destructor.
     */
    virtual ~Test();

    /**
     * Starts a test.
     */
    virtual void run() = 0;

protected:
    typedef void (UserModel::*stimulus_t)(Process &process, Interface &iface, Message &message);

    /**
     * Initializes a test.
     */
    void initialize();

    /**
     * Finalizes a test.
     */
    void finalize();

    /**
     * Starts a process for stimulus processing.
     */
    pid_t start(stimulus_t, Interface &iface, Message &message);

    /**
     * Emulates a cycle.
     */
    void cycle();

    /**
     * Emulates n cycles.
     */
    void delay(unsigned n);

    /// Model or mediator under test.
    UserModel &model;
};

template<class UserModel>
Test<UserModel>::Test(UserModel &model_under_test): model(model_under_test)
{
}

template<class UserModel>
Test<UserModel>::~Test()
{
}

template<class UserModel>
void Test<UserModel>::initialize()
{
    model.initialize();
}

template<class UserModel>
void Test<UserModel>::finalize()
{
    model.finalize();
}

template<class UserModel>
pid_t Test<UserModel>::start(stimulus_t stimulus, Interface &iface, Message &message)
{
    model.start(stimulus, iface, message);
}

template<class UserModel>
void Test<UserModel>::cycle()
{
    model.cycle();
}

template<class UserModel>
void Test<UserModel>::delay(unsigned n)
{
    model.delay(n);
}

}} // namespace cpptesk::hw

