/*
 * Copyright 2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* 
 * File: aspect_coverage.c
 * Description: 
 *       This file contains implementation of Aspectrace aspect "coverage".
 */


#include <utils/assertion.h>
#include "aspect_coverage.h"
#include "aspectrace_internals.h"

static const char ASPECT[] = "coverage";


struct aspect_coverage_element  {
	const char *name;
	struct aspect_coverage_element *firstChild;
	struct aspect_coverage_element *nextSibling;
};

aspect_coverage_element *create_CoverageElement(const char *name, aspect_coverage_element *children, ...)
{
/* !!! WARNING !!! Unsafe pointer conversion. Parameters are treated as array.
 * Usually it works. */
	return create_CoverageElement_a(name, &children);
}

aspect_coverage_element *create_CoverageElement_a(const char *name, aspect_coverage_element **children)
{
aspect_coverage_element *res = malloc( sizeof(aspect_coverage_element) );
aspect_coverage_element *cur;
int ctr = 0;

	assertion( name != NULL, FORMAT("create_CoverageElement: NULL name") );
	assertion( res != NULL, FORMAT("create_CoverageElement: No memory") );

#ifdef ALLOC_NEW
	res->name = strdup(name);
#else
	res->name = name;
#endif

	res->nextSibling = NULL;

	res->firstChild = cur = children[0];
	while (cur != NULL) {
		cur->nextSibling = children[ ctr++ ];
		cur = cur->nextSibling;
	}

	return res;
}

/** @return next sibling */
static aspect_coverage_element * destroy_CoverageElement(aspect_coverage_element *elem)
{
aspect_coverage_element *next;

#ifdef ALLOC_NEW
	free( elem->name );
#endif

	while (elem->firstChild != NULL) {
		elem->firstChild = destroy_CoverageElement( elem->firstChild );
	}
	
	next = elem->nextSibling;  /* will be cleared by parent */
	free(elem);
	return next;
}


void aspect_coverage_traceCoverageElement(
		const char *aspect,
		const char *structName,
		const char *elems, ...)
{
/* !!! WARNING !!! Unsafe pointer conversion. Parameters are treated as array.
 * Usually it works. */
	aspect_coverage_traceCoverageElement_a(aspect, structName, &elems);
}

void aspect_coverage_traceCoverageElement_a(
		const char *aspect,
		const char *structName,
		const char **elems)
{
TraceNode *curNode, *rootNode;
int ctr = 1;

	if (!aspectrace_allowed(ASPECT, "element"))
		return;

	assertion( elems != NULL && elems[0] != NULL,
		FORMAT("aspect_coverage_traceCoverageElement: Empty name chain") );

	{
		const char *rootAttrs[] = { "aspect", aspect, "cs", structName, "name", elems[0], NULL };
		rootNode = curNode = create_TraceNode( "element", rootAttrs, NULL );
	}

	for (;;) {
		const char *nextName = elems[ ctr++ ];
		const char *attrs[] = { "name", nextName, NULL };
		if (nextName == NULL)	break;
		curNode->firstChild = create_TraceNode( "element", attrs, NULL );
		curNode = curNode->firstChild;
	}

	aspect_trace( ASPECT, NULL, rootNode );
}


static TraceNode *convert_CoverageElement_to_TraceNode(aspect_coverage_element *ce)
{
TraceNode *res, **childrenNodes;
const char *attrs[] = { "name", ce->name, NULL };
aspect_coverage_element *curChild;
int i;

	curChild = ce->firstChild;
	for (i = 0; curChild != NULL; i++)
		curChild = curChild->nextSibling;

	childrenNodes = (TraceNode**)calloc( i+1, sizeof(TraceNode*) );
	assertion( childrenNodes != 0, FORMAT("convert_CoverageElement_to_TraceNode: No memory") );
	childrenNodes[i] = NULL;

	curChild = ce->firstChild;
	for (i = 0; curChild != NULL; i++) {
		aspect_coverage_element *nextChild = curChild->nextSibling;
		childrenNodes[i] = convert_CoverageElement_to_TraceNode( curChild );	/* destroys curChild */
		curChild = nextChild;
	}

	res = create_TraceNode_a("element", attrs, childrenNodes);

	/*** Cleanup ***/

	/* All children are already destroyed by convert_CoverageElement_to_TraceNode().
	 * Do not let function destroy_CoverageElement(ce) access them. */
	ce->firstChild = NULL;

	free( childrenNodes );
	destroy_CoverageElement( ce );

	return res;
}

void aspect_coverage_traceCoverageStructure(
		const char *aspect,
		const char *name,
		aspect_coverage_element *elems, ...)
{
/* !!! WARNING !!! Unsafe pointer conversion. Parameters are treated as array.
 * Usually it works. */
	aspect_coverage_traceCoverageStructure_a(aspect, name, &elems);
}

void aspect_coverage_traceCoverageStructure_a(
		const char *aspect,
		const char *name,
		aspect_coverage_element **elems)
{
TraceNode **elemNodes;
int i, ctr = 0;

	if (!aspectrace_allowed(ASPECT, "structure"))
		return;

	assertion( elems != NULL && elems[0] != NULL,
		FORMAT("aspect_coverage_traceCoverageStructure: Empty coverage structure") );
	assertion( aspect != NULL && name != NULL,
		FORMAT("aspect_coverage_traceCoverageStructure: NULL parameter") );

	/* Count coverage elements */
	for (;;) {
		aspect_coverage_element *ce = elems[ ctr++ ];
		if (ce == NULL)	break;
	}

	elemNodes = (TraceNode **)calloc( ctr, sizeof(TraceNode*) );
	assertion( elemNodes != 0, FORMAT("aspect_coverage_traceCoverageStructure: No memory") );
	ctr--;

	/* Create NULL-terminated array of TraceNode's */
	for (i = 0; i < ctr; i++) {
		elemNodes[i] = convert_CoverageElement_to_TraceNode( elems[i] );
	}
	elemNodes[ctr] = NULL;

	// Create and trace message
	{
		const char *rootAttrs[] = { "aspect", aspect, "name", name, NULL };
		aspect_trace( ASPECT, NULL,
			create_TraceNode_a("structure", rootAttrs, elemNodes)
		);
	}

	free( elemNodes );
}

void aspect_coverage_traceCteskCS(CoverageDesc *cs)
{
int i;
TraceNode **children;

	if (!aspectrace_allowed(ASPECT, "structure"))
		return;

	assertion( cs != NULL, FORMAT("aspect_coverage_traceCteskCS: NULL argument") );
	assertion( cs->kind == PRIMARY_COVERAGE,
		FORMAT("aspect_coverage_traceCteskCS: derived coverage structures not supported") );

	children = (TraceNode**)calloc( cs->size+1, sizeof(TraceNode*) );
	assertion( children != NULL, FORMAT("aspect_coverage_traceCteskCS: No memory") );

	for (i = 0; i < cs->size; i++) {
		const char *attrs[] = { "name", cs->data[i], NULL };
		children[i] = create_TraceNode("element", attrs, NULL);
	}
	children[cs->size] = NULL;

	{
		const char *rootAttrs[] = { "aspect", "oracle", "name", cs->name, NULL };
		aspect_trace(ASPECT, NULL, create_TraceNode_a("structure", rootAttrs, children));
	}

	free( children );
}

